# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/base/collection/indexed_dict.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1562 bytes
from collections import OrderedDict


class IndexedDict(OrderedDict):
    def __init__(self, *args, **kwds):
        self._IndexedDict__keys = []
        (super().__init__)(*args, **kwds)

    def __setitem__(self, key, value, *args, **kwds):
        (super().__setitem__)(key, value, *args, **kwds)
        if key not in self._IndexedDict__keys:
            self._IndexedDict__keys.append(key)

    def __delitem__(self, key, *args, **kwds):
        (super().__delitem__)(key, *args, **kwds)
        self._IndexedDict__keys.remove(key)

    def clear(self):
        super().clear()
        self._IndexedDict__keys = []

    def popitem(self, last=True):
        item = super().popitem(last)
        self._IndexedDict__keys.pop(-1 if last else 0)
        return item

    def keys(self):
        return self._IndexedDict__keys

    def item_by_index(self, ix):
        key = self._IndexedDict__keys[ix]
        return (key, self[key])

    def key_by_index(self, ix):
        return self._IndexedDict__keys[ix]

    def value_by_index(self, ix):
        return self[self._IndexedDict__keys[ix]]

    def index_by_key(self, key):
        return self._IndexedDict__keys.index(key)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/playhead.py
# Compiled at: 2023-12-18 13:26:15
# Size of source mod 2**32: 3810 bytes
from ...base import depends, listens
from ...live import liveobj_valid
from .. import Component
from . import DEFAULT_STEP_TRANSLATION_CHANNEL


class PlayheadComponent(Component):
    @depends(playhead=None, sequencer_clip=None, grid_resolution=None)
    def __init__(
        self,
        name="Playhead",
        playhead=None,
        sequencer_clip=None,
        grid_resolution=None,
        paginator=None,
        notes=None,
        triplet_notes=None,
        channels=None,
        *a,
        **k,
    ):
        (super().__init__)(a, name=name, **k)
        self._notes = notes or tuple(range(8))
        self._triplet_notes = triplet_notes or tuple(range(6))
        self._channels = channels or [DEFAULT_STEP_TRANSLATION_CHANNEL]
        self._playhead = playhead
        self._sequencer_clip = sequencer_clip
        self._grid_resolution = grid_resolution
        self._paginator = paginator
        self.register_slot(self._grid_resolution, self.update, "index")
        self.register_slot(self._paginator, self.update, "page")
        self.register_slot(self.song, self.update, "is_playing")
        self._PlayheadComponent__on_sequencer_clip_playing_status_changed.subject = sequencer_clip
        self._PlayheadComponent__on_sequencer_clip_changed()

    def update(self):
        super().update()
        playhead_clip = None
        sequencer_clip = self._sequencer_clip.clip
        if self.is_enabled():
            if not liveobj_valid(sequencer_clip) or self.song.is_playing:
                if sequencer_clip.is_arrangement_clip or sequencer_clip.is_playing:
                    playhead_clip = sequencer_clip
                self._playhead.clip = playhead_clip
                self._playhead.set_feedback_channels(self._channels)
                if playhead_clip:
                    self._update_playhead_notes()
                    self._playhead.start_time = self._paginator.page_time
                    self._playhead.step_length = self._grid_resolution.step_length

    def _update_playhead_notes(self):
        self._playhead.notes = list(self._triplet_notes if self._grid_resolution.is_triplet else self._notes)

    @listens("clip")
    def __on_sequencer_clip_changed(self):
        self.update()

    @listens("clip.playing_status")
    def __on_sequencer_clip_playing_status_changed(self):
        self.update()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/live/action.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 11985 bytes
import logging
from functools import singledispatch, wraps
from sys import maxsize
from typing import Optional, Union

from Live import Clip, ClipSlot, DeviceParameter, Scene, Track
from Live.Base import LimitationError
from Live.Song import Quantization

from . import liveobj_changed, liveobj_valid, scene_index, song, track_index
from .util import is_clip_new_recording, raise_type_error_for_liveobj

logger = logging.getLogger(__name__)


def action(func):
    singledispatch_func = singledispatch(func)

    @wraps(func)
    def wrapper(target, *a, **k):
        if liveobj_valid(target):
            try:
                return singledispatch_func(target, *a, **k)
            except (AttributeError, AssertionError, ValueError, RuntimeError, TypeError, LimitationError) as e:
                try:
                    logger.debug("An exception occurred when attempting to perform an action: %s", e)
                finally:
                    e = None
                    del e

            return False
        return None

    wrapper.register = singledispatch_func.register
    return wrapper


@action
def arm(track: Track, exclusive=None) -> bool:
    if track.can_be_armed:
        exclusive = exclusive if exclusive is not None else song().exclusive_arm
        new_value = not track.arm
        for t in song().tracks:
            if t.can_be_armed:
                if not (t == track or track).is_part_of_selection or t.is_part_of_selection:
                    t.arm = new_value
                else:
                    if exclusive:
                        if t.arm:
                            t.arm = False

        return True
    return False


@action
def delete(deletable: Union[(Clip, ClipSlot, Scene, Track)]) -> bool:
    return raise_type_error_for_liveobj(deletable)


@delete.register
def _(deletable: Clip):
    if deletable.is_arrangement_clip:
        deletable.canonical_parent.delete_clip(deletable)
    else:
        deletable.canonical_parent.delete_clip()
    return True


@delete.register
def _(deletable: ClipSlot):
    deletable.delete_clip()
    return True


@delete.register
def _(deletable: Scene):
    song().delete_scene(scene_index(deletable))
    return True


@delete.register
def _(deletable: Track):
    if deletable in song().return_tracks:
        song().delete_return_track(track_index(deletable, track_list=(list(song().return_tracks))))
    else:
        song().delete_track(track_index(deletable))
    return True


@action
def delete_notes_with_pitch(clip: Clip, pitch: int) -> bool:
    args = {"from_time": 0, "from_pitch": pitch, "time_span": maxsize, "pitch_span": 1}
    if (clip.get_notes_extended)(**args):
        (clip.remove_notes_extended)(**args)
        return True
    return False


@action
def delete_notes_in_range(clip, from_time, time_span):
    args = {"from_time": from_time, "from_pitch": 0, "time_span": time_span, "pitch_span": 128}
    if (clip.get_notes_extended)(**args):
        (clip.remove_notes_extended)(**args)
        return True
    return False


@action
def duplicate_loop(clip: Clip) -> bool:
    clip.duplicate_loop()
    clip.view.show_loop()
    return True


@action
def duplicate(duplicatable: Union[(Clip, ClipSlot, Scene, Track)]) -> bool:
    return raise_type_error_for_liveobj(duplicatable)


@duplicate.register
def _(duplicatable: Clip):
    if duplicatable.is_arrangement_clip:
        track = duplicatable.canonical_parent
        track.duplicate_clip_to_arrangement(duplicatable, duplicatable.end_time)
        return True
    return duplicate(duplicatable.canonical_parent)


@duplicate.register
def _(duplicatable: ClipSlot):
    track = duplicatable.canonical_parent
    track.duplicate_clip_slot(list(track.clip_slots).index(duplicatable))
    return True


@duplicate.register
def _(duplicatable: Scene):
    song().duplicate_scene(scene_index(duplicatable))
    return True


@duplicate.register
def _(duplicatable: Track):
    song().duplicate_track(track_index(duplicatable))
    return True


@action
def duplicate_clip_special(clip: Clip) -> bool:
    if clip.is_arrangement_clip:
        track = clip.canonical_parent
        song().view.detail_clip = track.duplicate_clip_to_arrangement(clip, clip.end_time)
    else:
        slot = clip.canonical_parent
        track = slot.canonical_parent
        slot_index = list(track.clip_slots).index(slot)
        next_slot_index = track.duplicate_clip_slot(slot_index)
        song().view.selected_scene = song().scenes[next_slot_index]
        if clip.is_playing:
            track.clip_slots[next_slot_index].fire(force_legato=True, launch_quantization=(Quantization.q_no_q))
    return True


@action
def select(selectable: Union[(Clip, ClipSlot, Scene, Track)]) -> bool:
    return raise_type_error_for_liveobj(selectable)


@select.register
def _(selectable: Clip):
    return select(selectable.canonical_parent)


@select.register
def _(selectable: ClipSlot):
    if not liveobj_changed(song().view.highlighted_clip_slot, selectable):
        return False
    song().view.highlighted_clip_slot = selectable
    return True


@select.register
def _(selectable: Scene):
    if not liveobj_changed(song().view.selected_scene, selectable):
        return False
    song().view.selected_scene = selectable
    return True


@select.register
def _(selectable: Track):
    if not liveobj_changed(song().view.selected_track, selectable):
        return False
    song().view.selected_track = selectable
    return True


@action
def fire(fireable: Union[(Clip, ClipSlot, Scene)], button_state=None) -> bool:
    if isinstance(fireable, ClipSlot):
        if fireable.has_clip:
            fireable = fireable.clip
    if button_state is None:
        fireable.fire()
    else:
        fireable.set_fire_button_state(button_state)
        if button_state:
            if song().select_on_launch:
                select(fireable)
    return True


@action
def toggle_or_cycle_parameter_value(parameter: DeviceParameter) -> bool:
    if parameter.is_quantized:
        if parameter.value + 1 > parameter.max:
            parameter.value = parameter.min
        else:
            parameter.value = parameter.value + 1
    else:
        parameter.value = parameter.max if parameter.value == parameter.min else parameter.min
    return True


@action
def set_loop_start(clip: Clip, loop_start: float, show_loop: Optional[bool] = True) -> bool:
    if is_clip_new_recording(clip):
        return False
    clip.loop_start = loop_start
    clip.start_marker = loop_start
    if show_loop:
        clip.view.show_loop()
    return True


@action
def set_loop_end(clip: Clip, loop_end: float, show_loop: Optional[bool] = True) -> bool:
    if is_clip_new_recording(clip):
        return False
    clip.loop_end = loop_end
    clip.end_marker = loop_end
    if show_loop:
        clip.view.show_loop()
    return True


@action
def set_loop_position(clip, loop_start, loop_end):
    if is_clip_new_recording(clip):
        return False
    if loop_start < clip.loop_end:
        set_loop_start(clip, loop_start, show_loop=False)
        set_loop_end(clip, loop_end, show_loop=False)
    else:
        set_loop_end(clip, loop_end, show_loop=False)
        set_loop_start(clip, loop_start, show_loop=False)
    clip.view.show_loop()
    return True



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/elements/sysex.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 6400 bytes
from contextlib import contextmanager
from inspect import signature

from ...base import index_if, lazy_attribute
from .. import MIDI_SYSEX_TYPE, InputControlElement


class SysexElement(InputControlElement):
    def __init__(
        self,
        send_message_generator=None,
        enquire_message=None,
        default_value=None,
        optimized=False,
        use_first_byte_as_value=False,
        *a,
        **k,
    ):
        (super().__init__)(a, msg_type=MIDI_SYSEX_TYPE, send_should_depend_on_forwarding=False, **k)
        self._send_message_generator = send_message_generator
        self._enquire_message = enquire_message
        self._default_value = default_value
        self._optimized = optimized
        self._use_first_byte_as_value = use_first_byte_as_value
        self._is_deferring_send = False
        self._deferred_message = None

    def message_map_mode(self):
        raise AssertionError("SysexElement doesn't support mapping.")

    def receive_value(self, value):
        if self._use_first_byte_as_value:
            value = value[0]
        super().receive_value(value)

    def send_value(self, *a, **k):
        if self._send_message_generator:
            message = (self._send_message_generator)(*a, **k)
            if message is not None:
                self._do_send_value(message)

    def enquire_value(self):
        self.send_midi(self._enquire_message)

    def reset(self):
        if self._default_value is not None:
            self.send_value(self._default_value)

    @contextmanager
    def deferring_send(self):
        try:
            self._deferred_message = None
            self._is_deferring_send = True
            yield
        finally:
            self._is_deferring_send = False
            if self._deferred_message is not None:
                self._do_send_value(self._deferred_message)

    def _do_send_value(self, message):
        if not (self._is_deferring_send or self)._optimized or message != self._last_sent_message:
            if self.send_midi(message):
                self._last_sent_message = message
        else:
            if self._is_deferring_send:
                self._deferred_message = message

    @property
    def _last_sent_value(self):
        if self._last_sent_message:
            return self._last_sent_message
        return -1


class CachingSendMessageGenerator:
    def __init__(self, generator):
        self._generator = generator
        self._generator_params = signature(self._generator).parameters

    def __call__(self, *a, **k):
        message = None
        if k:
            self._update_cached_values_from_kwargs(k)
            if None not in self._cached_argument_values:
                message = (self._generator)(*self._cached_argument_values)
        else:
            for i, value in enumerate(a):
                self._cached_argument_values[i] = value

            message = (self._generator)(*a)
        return message

    @lazy_attribute
    def _cached_argument_values(self):
        return [None] * len(self._generator_params)

    def _update_cached_values_from_kwargs(self, kwargs):
        for key, value in kwargs.items():
            kwarg_index = index_if(lambda param_name, kwarg_name=key: param_name == kwarg_name, self._generator_params)
            self._cached_argument_values[kwarg_index] = value
